@@ -1,5 +1,6 @@
 """Conversions between rotation representations."""
 import math
+import warnings
 import numpy as np
 from ._utils import (
     check_matrix, check_quaternion, check_axis_angle, check_compact_axis_angle,
@@ -297,7 +298,7 @@ def active_matrix_from_extrinsic_euler_xzx(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -317,7 +318,7 @@ def active_matrix_from_intrinsic_euler_xyx(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -337,7 +338,7 @@ def active_matrix_from_extrinsic_euler_xyx(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -357,7 +358,7 @@ def active_matrix_from_intrinsic_euler_yxy(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -377,7 +378,7 @@ def active_matrix_from_extrinsic_euler_yxy(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -397,7 +398,7 @@ def active_matrix_from_intrinsic_euler_yzy(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -417,7 +418,7 @@ def active_matrix_from_extrinsic_euler_yzy(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -437,7 +438,7 @@ def active_matrix_from_intrinsic_euler_zyz(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -463,7 +464,7 @@ def active_matrix_from_extrinsic_euler_zyz(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -483,7 +484,7 @@ def active_matrix_from_intrinsic_euler_zxz(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -509,7 +510,7 @@ def active_matrix_from_extrinsic_euler_zxz(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -529,7 +530,7 @@ def active_matrix_from_intrinsic_euler_xzy(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -549,7 +550,7 @@ def active_matrix_from_extrinsic_euler_xzy(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -569,7 +570,7 @@ def active_matrix_from_intrinsic_euler_xyz(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -589,7 +590,7 @@ def active_matrix_from_extrinsic_euler_xyz(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -609,7 +610,7 @@ def active_matrix_from_intrinsic_euler_yxz(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -629,7 +630,7 @@ def active_matrix_from_extrinsic_euler_yxz(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -649,7 +650,7 @@ def active_matrix_from_intrinsic_euler_yzx(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -669,7 +670,7 @@ def active_matrix_from_extrinsic_euler_yzx(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -689,7 +690,7 @@ def active_matrix_from_intrinsic_euler_zyx(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -709,7 +710,7 @@ def active_matrix_from_extrinsic_euler_zyx(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -729,7 +730,7 @@ def active_matrix_from_intrinsic_euler_zxy(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -749,7 +750,7 @@ def active_matrix_from_extrinsic_euler_zxy(e):
 
     Returns
     -------
-    R : array-like, shape (3, 3)
+    R : array, shape (3, 3)
         Rotation matrix
     """
     alpha, beta, gamma = e
@@ -776,6 +777,73 @@ def active_matrix_from_extrinsic_roll_pitch_yaw(rpy):
     return active_matrix_from_extrinsic_euler_xyz(rpy)
 
 
+def check_axis_index(name, i):
+    """Checks axis index.
+
+    Parameters
+    ----------
+    name : str
+        Name of the axis. Required for the error message.
+
+    i : int from [0, 1, 2]
+        Index of the axis (0: x, 1: y, 2: z)
+
+    Raises
+    ------
+    ValueError
+        If basis is invalid
+    """
+    if i not in [0, 1, 2]:
+        raise ValueError("Axis index %s (%d) must be in [0, 1, 2]" % (name, i))
+
+
+def matrix_from_euler(e, i, j, k, extrinsic):
+    """General method to compute active rotation matrix from any Euler angles.
+
+    Parameters
+    ----------
+    e : array-like, shape (3,)
+        Extracted rotation angles in radians about the axes i, j, k in this
+        order. The first and last angle are normalized to [-pi, pi]. The middle
+        angle is normalized to either [0, pi] (proper Euler angles) or
+        [-pi/2, pi/2] (Cardan / Tait-Bryan angles).
+
+    i : int from [0, 1, 2]
+        The first rotation axis (0: x, 1: y, 2: z)
+
+    j : int from [0, 1, 2]
+        The second rotation axis (0: x, 1: y, 2: z)
+
+    k : int from [0, 1, 2]
+        The third rotation axis (0: x, 1: y, 2: z)
+
+    extrinsic : bool
+        Do we use extrinsic transformations? Intrinsic otherwise.
+
+    Returns
+    -------
+    R : array, shape (3, 3)
+        Active rotation matrix
+
+    Raises
+    ------
+    ValueError
+        If basis is invalid
+    """
+    check_axis_index("i", i)
+    check_axis_index("j", j)
+    check_axis_index("k", k)
+
+    alpha, beta, gamma = e
+    if not extrinsic:
+        i, k = k, i
+        alpha, gamma = gamma, alpha
+    R = active_matrix_from_angle(k, gamma).dot(
+        active_matrix_from_angle(j, beta)).dot(
+        active_matrix_from_angle(i, alpha))
+    return R
+
+
 def _general_intrinsic_euler_from_active_matrix(
         R, n1, n2, n3, proper_euler, strict_check=True):
     """General algorithm to extract intrinsic euler angles from a matrix.
@@ -901,7 +969,7 @@ def intrinsic_euler_xzx_from_active_matrix(R, strict_check=True):
 
 
 def extrinsic_euler_xzx_from_active_matrix(R, strict_check=True):
-    """Compute active rotation matrix from extrinsic xzx Euler angles.
+    """Compute extrinsic xzx Euler angles from active rotation matrix.
 
     Parameters
     ----------
@@ -914,7 +982,7 @@ def extrinsic_euler_xzx_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around x-, z-, and x-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -977,7 +1045,7 @@ def intrinsic_euler_yxy_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around y-, x'-, and y''-axes (intrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -998,7 +1066,7 @@ def extrinsic_euler_yxy_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around y-, x-, and y-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1019,7 +1087,7 @@ def intrinsic_euler_yzy_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around y-, z'-, and y''-axes (intrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1040,7 +1108,7 @@ def extrinsic_euler_yzy_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around y-, z-, and y-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1061,7 +1129,7 @@ def intrinsic_euler_zyz_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around z-, y'-, and z''-axes (intrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1082,7 +1150,7 @@ def extrinsic_euler_zyz_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around z-, y-, and z-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1103,7 +1171,7 @@ def intrinsic_euler_zxz_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around z-, x'-, and z''-axes (intrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1124,7 +1192,7 @@ def extrinsic_euler_zxz_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around z-, x-, and z-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1145,7 +1213,7 @@ def intrinsic_euler_xzy_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around x-, z'-, and y''-axes (intrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1166,7 +1234,7 @@ def extrinsic_euler_xzy_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around x-, z-, and y-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1187,7 +1255,7 @@ def intrinsic_euler_xyz_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around x-, y'-, and z''-axes (intrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1208,7 +1276,7 @@ def extrinsic_euler_xyz_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around x-, y-, and z-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1229,7 +1297,7 @@ def intrinsic_euler_yxz_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around y-, x'-, and z''-axes (intrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1250,7 +1318,7 @@ def extrinsic_euler_yxz_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around y-, x-, and z-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1271,7 +1339,7 @@ def intrinsic_euler_yzx_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around y-, z'-, and x''-axes (intrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1292,7 +1360,7 @@ def extrinsic_euler_yzx_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around y-, z-, and x-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
@@ -1376,13 +1444,196 @@ def extrinsic_euler_zxy_from_active_matrix(R, strict_check=True):
 
     Returns
     -------
-    e : array-like, shape (3,)
+    e : array, shape (3,)
         Angles for rotation around z-, x-, and y-axes (extrinsic rotations)
     """
     return _general_intrinsic_euler_from_active_matrix(
         R, unity, unitx, unitz, False, strict_check)[::-1]
 
 
+def euler_from_matrix(R, i, j, k, extrinsic, strict_check=True):
+    """General method to extract any Euler angles from active rotation matrix.
+
+    Parameters
+    ----------
+    R : array-like, shape (3, 3)
+        Active rotation matrix
+
+    i : int from [0, 1, 2]
+        The first rotation axis (0: x, 1: y, 2: z)
+
+    j : int from [0, 1, 2]
+        The second rotation axis (0: x, 1: y, 2: z)
+
+    k : int from [0, 1, 2]
+        The third rotation axis (0: x, 1: y, 2: z)
+
+    extrinsic : bool
+        Do we use extrinsic transformations? Intrinsic otherwise.
+
+    strict_check : bool, optional (default: True)
+        Raise a ValueError if the rotation matrix is not numerically close
+        enough to a real rotation matrix. Otherwise we print a warning.
+
+    Returns
+    -------
+    euler_angles : array, shape (3,)
+        Extracted rotation angles in radians about the axes i, j, k in this
+        order. The first and last angle are normalized to [-pi, pi]. The middle
+        angle is normalized to either [0, pi] (proper Euler angles) or
+        [-pi/2, pi/2] (Cardan / Tait-Bryan angles).
+
+    Raises
+    ------
+    ValueError
+        If basis is invalid
+
+    References
+    ----------
+    Shuster, Markley: General Formula for Extracting the Euler Angles,
+    https://arc.aiaa.org/doi/abs/10.2514/1.16622
+    """
+    check_axis_index("i", i)
+    check_axis_index("j", j)
+    check_axis_index("k", k)
+
+    basis_vectors = [unitx, unity, unitz]
+    proper_euler = i == k
+    if extrinsic:
+        i, k = k, i
+    e = _general_intrinsic_euler_from_active_matrix(
+        R, basis_vectors[i], basis_vectors[j], basis_vectors[k], proper_euler,
+        strict_check)
+
+    if extrinsic:
+        e = e[::-1]
+
+    return e
+
+
+def euler_from_quaternion(q, i, j, k, extrinsic):
+    """General method to extract any Euler angles from quaternions.
+
+    Parameters
+    ----------
+    q : array-like, shape (4,)
+        Unit quaternion to represent rotation: (w, x, y, z)
+
+    i : int from [0, 1, 2]
+        The first rotation axis (0: x, 1: y, 2: z)
+
+    j : int from [0, 1, 2]
+        The second rotation axis (0: x, 1: y, 2: z)
+
+    k : int from [0, 1, 2]
+        The third rotation axis (0: x, 1: y, 2: z)
+
+    extrinsic : bool
+        Do we use extrinsic transformations? Intrinsic otherwise.
+
+    Returns
+    -------
+    euler_angles : array, shape (3,)
+        Extracted rotation angles in radians about the axes i, j, k in this
+        order. The first and last angle are normalized to [-pi, pi]. The middle
+        angle is normalized to either [0, pi] (proper Euler angles) or
+        [-pi/2, pi/2] (Cardan / Tait-Bryan angles).
+
+    Raises
+    ------
+    ValueError
+        If basis is invalid
+
+    References
+    ----------
+    Bernardes, Evandro; Viollet, Stephane: Quaternion to Euler angles
+    conversion: A direct, general and computationally efficient method,
+    https://doi.org/10.1371/journal.pone.0276302
+    """
+    q = check_quaternion(q)
+
+    check_axis_index("i", i)
+    check_axis_index("j", j)
+    check_axis_index("k", k)
+
+    i += 1
+    j += 1
+    k += 1
+
+    # The original algorithm assumes extrinsic convention. Hence, we swap
+    # the order of axes for intrinsic rotation.
+    if not extrinsic:
+        i, k = k, i
+
+    # Proper Euler angles rotate about the same axis in the first and last
+    # rotation. If this is not the case, they are called Cardan or Tait-Bryan
+    # angles and have to be handled differently.
+    proper_euler = i == k
+    if proper_euler:
+        k = 6 - i - j
+
+    sign = (i - j) * (j - k) * (k - i) // 2
+    a = q[0]
+    b = q[i]
+    c = q[j]
+    d = q[k] * sign
+
+    if not proper_euler:
+        a, b, c, d = a - c, b + d, c + a, d - b
+
+    # Equation 34 is used instead of Equation 35 as atan2 it is numerically
+    # more accurate than acos.
+    angle_j = 2.0 * math.atan2(math.hypot(c, d),
+                               math.hypot(a, b))
+
+    # Check for singularities
+    if abs(angle_j) <= eps:
+        singularity = 1
+    elif abs(angle_j - math.pi) <= eps:
+        singularity = 2
+    else:
+        singularity = 0
+
+    # Equation 25
+    # (theta_1 + theta_3) / 2
+    half_sum = math.atan2(b, a)
+    # (theta_1 - theta_3) / 2
+    half_diff = math.atan2(d, c)
+
+    if singularity == 0:  # no singularity
+        # Equation 32
+        angle_i = half_sum + half_diff
+        angle_k = half_sum - half_diff
+    elif extrinsic:  # singularity
+        angle_k = 0.0
+        if singularity == 1:
+            angle_i = 2.0 * half_sum
+        else:
+            assert singularity == 2
+            angle_i = 2.0 * half_diff
+    else:  # intrinsic, singularity
+        angle_i = 0.0
+        if singularity == 1:
+            angle_k = 2.0 * half_sum
+        else:
+            assert singularity == 2
+            angle_k = -2.0 * half_diff
+
+    if not proper_euler:
+        # Equation 43
+        angle_j -= math.pi / 2.0
+        # Equation 44
+        angle_i *= sign
+
+    angle_k = norm_angle(angle_k)
+    angle_i = norm_angle(angle_i)
+
+    if extrinsic:
+        return np.array([angle_k, angle_j, angle_i])
+
+    return np.array([angle_i, angle_j, angle_k])
+
+
 def axis_angle_from_matrix(R, strict_check=True, check=True):
     """Compute axis-angle from rotation matrix.
 
@@ -1672,6 +1923,43 @@ def quaternion_from_matrix(R, strict_check=True):
     return q
 
 
+def quaternion_from_angle(basis, angle):
+    """Compute quaternion from rotation about basis vector.
+
+    Parameters
+    ----------
+    basis : int from [0, 1, 2]
+        The rotation axis (0: x, 1: y, 2: z)
+
+    angle : float
+        Rotation angle
+
+    Returns
+    -------
+    q : array, shape (4,)
+        Unit quaternion to represent rotation: (w, x, y, z)
+
+    Raises
+    ------
+    ValueError
+        If basis is invalid
+    """
+    half_angle = 0.5 * angle
+    c = math.cos(half_angle)
+    s = math.sin(half_angle)
+
+    if basis == 0:
+        q = np.array([c, s, 0.0, 0.0])
+    elif basis == 1:
+        q = np.array([c, 0.0, s, 0.0])
+    elif basis == 2:
+        q = np.array([c, 0.0, 0.0, s])
+    else:
+        raise ValueError("Basis must be in [0, 1, 2]")
+
+    return q
+
+
 def quaternion_from_axis_angle(a):
     """Compute quaternion from axis-angle.
 
@@ -1686,7 +1974,7 @@ def quaternion_from_axis_angle(a):
 
     Returns
     -------
-    q : array-like, shape (4,)
+    q : array, shape (4,)
         Unit quaternion to represent rotation: (w, x, y, z)
     """
     a = check_axis_angle(a)
@@ -1710,7 +1998,7 @@ def quaternion_from_compact_axis_angle(a):
 
     Returns
     -------
-    q : array-like, shape (4,)
+    q : array, shape (4,)
         Unit quaternion to represent rotation: (w, x, y, z)
     """
     a = axis_angle_from_compact_axis_angle(a)
@@ -1727,7 +2015,7 @@ def quaternion_xyzw_from_wxyz(q_wxyz):
 
     Returns
     -------
-    q_xyzw : array-like, shape (4,)
+    q_xyzw : array, shape (4,)
         Quaternion with scalar part after vector part
     """
     q_wxyz = check_quaternion(q_wxyz)
@@ -1744,7 +2032,7 @@ def quaternion_wxyz_from_xyzw(q_xyzw):
 
     Returns
     -------
-    q_wxyz : array-like, shape (4,)
+    q_wxyz : array, shape (4,)
         Quaternion with scalar part before vector part
     """
     q_xyzw = check_quaternion(q_xyzw)
@@ -1761,8 +2049,11 @@ def quaternion_from_extrinsic_euler_xyz(e):
 
     Returns
     -------
-    q : array-like, shape (4,)
+    q : array, shape (4,)
         Unit quaternion to represent rotation: (w, x, y, z)
     """
+    warnings.warn(
+        "quaternion_from_extrinsic_euler_xyz is deprecated, use "
+        "quaternion_from_euler", DeprecationWarning)
     R = active_matrix_from_extrinsic_euler_xyz(e)
     return quaternion_from_matrix(R)
