@@ -38,6 +38,13 @@ class TransformManager(object):
         for large graphs. You can check the consistency explicitly with
         :func:`TransformManager.check_consistency`.
 
+    The TransformManager does not directly support serialization because
+    we don't want to decide for a specific format. However, it allows
+    conversion to a dict with only primitive types that is serializable,
+    for instance, as JSON. If a more compact format is required, binary
+    formats like msgpack can be used. Use :func:`TransformManager.to_dict`
+    and :func:`TransformManager.from_dict` for this purpose.
+
     Parameters
     ----------
     strict_check : bool, optional (default: True)
@@ -62,13 +69,13 @@ class TransformManager(object):
         # We have to store the index n associated to a transformation to be
         # able to remove the transformation later
         self.transform_to_ij_index = {}
-        # Same information as sparse matrix
+        # Connection information as sparse matrix
         self.connections = sp.csr_matrix((0, 0))
 
         # Result of shortest path algorithm:
         # distance matrix (distance is the number of transformations)
         self.dist = np.empty(0)
-        self.predecessors = np.empty(0)
+        self.predecessors = np.empty(0, dtype=np.int32)
 
         self._cached_shortest_paths = {}
 
@@ -485,6 +492,79 @@ class TransformManager(object):
 
         graph.write_png(filename, prog=prog)
 
+    def to_dict(self):
+        """Convert the transform manager to a dict that is serializable.
+
+        Returns
+        -------
+        tm_dict : dict
+            Serializable dict.
+        """
+        return {
+            "class": self.__class__.__name__,
+            "strict_check": self.strict_check,
+            "check": self.check,
+            "transforms": [(k, v.ravel().tolist())
+                           for k, v in self.transforms.items()],
+            "nodes": self.nodes,
+            "i": self.i,
+            "j": self.j,
+            "transform_to_ij_index": list(self.transform_to_ij_index.items()),
+            "connections": {
+                "data": self.connections.data.tolist(),
+                "indices": self.connections.indices.tolist(),
+                "indptr": self.connections.indptr.tolist()
+            },
+            "dist": self.dist.tolist(),
+            "predecessors": self.predecessors.tolist()
+        }
+
+    @staticmethod
+    def from_dict(tm_dict):
+        """Create transform manager from dict.
+
+        Parameters
+        ----------
+        tm_dict : dict
+            Serializable dict.
+
+        Returns
+        -------
+        tm : TransformManager
+            Deserialized transform manager.
+        """
+        strict_check = tm_dict.get("strict_check")
+        check = tm_dict.get("check")
+        tm = TransformManager(strict_check=strict_check, check=check)
+        tm.set_transform_manager_state(tm_dict)
+        return tm
+
+    def set_transform_manager_state(self, tm_dict):
+        """Set state of transform manager from dict.
+
+        Parameters
+        ----------
+        tm_dict : dict
+            Serializable dict.
+        """
+        transforms = tm_dict.get("transforms")
+        self.transforms = dict([
+            (tuple(k), np.array(v).reshape(4, 4)) for k, v in transforms])
+        self.nodes = tm_dict.get("nodes")
+        self.i = tm_dict.get("i")
+        self.j = tm_dict.get("j")
+        self.transform_to_ij_index = dict(
+            [(tuple(k), v) for k, v in tm_dict.get("transform_to_ij_index")])
+        connections = tm_dict.get("connections")
+        self.connections = sp.csr_matrix((
+            connections["data"], connections["indices"],
+            connections["indptr"]))
+        n_nodes = len(self.nodes)
+        dist = np.array(tm_dict.get("dist"))
+        self.dist = dist.reshape(n_nodes, n_nodes)
+        predecessors = np.array(tm_dict.get("predecessors"), dtype=np.int32)
+        self.predecessors = predecessors.reshape(n_nodes, n_nodes)
+
 
 def _dot_display_name(name):  # pragma: no cover
     return name.replace("/", "")
